# core/bot.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2014 B.H.J Thate"

## IMPORTS

from core.utils import get_name, short_date
from core.errors import RemoteDisconnect
from core.dispatch import Dispatcher
from core.defines import BLA, ENDC
from core.object import Object

from core import kernel, fleet

import logging
import socket
import time

## BOTS

class Bot(Dispatcher):

    """ Base Bot class. """

    def __init__(zelf, *args, **kwargs):
        Dispatcher.__init__(zelf, *args, **kwargs)
        zelf.btype = get_name(zelf).split(".")[-1]
        zelf.channels = []
        zelf.waiting = True
        zelf._started = time.time()
        if "channel" in zelf: zelf.channels.append(zelf.channel)
        fleet.append(zelf)

    def _raw(zelf, *args, **kwargs): pass

    def join_channels(zelf, *args, **kwargs):
        """ join channels.  """ 
        for channel in zelf.channels: zelf.join(channel)

    def join(zelf, *args, **kwargs): pass

    def exit(zelf, *args, **kwargs):
        """ shutdown a bot. """
        zelf._status = ""

    def start(zelf, *args, **kwargs):
        """ start a bot. """
        zelf._status = "running"
        zelf.connect()
        while zelf._status:
             try: event = zelf.get_event()
             except RemoteDisconnect: break
             except EOFError: return
             if not zelf._status: break
             if event: zelf.put(zelf.handle_event, event)
             if zelf._status == "once": break
             time.sleep(0.05)
        zelf.ready()
 
    def get_prompt(zelf, *args, **kwargs):
        """ return a prompt. """
        return ""

    def get_event(zelf, *args, **kwargs):
        """ return an event. """
        o =  Object()
        o._target = zelf
        return o

    def connect(zelf, *args, **kwargs):
        """ connect to server. """
        pass

    def announce(zelf, *args, **kwargs):
        """ announce on channels. """
        for channel in zelf.channels: zelf.say(channel, args[0])
                   
    def say(zelf, *args, **kwargs): raise NotImplemented()

    def cmnd(zelf, *args, **kwargs):
        """ run a command with provided string. """
        o = Object()
        o._target = zelf
        o.txt = args[0]
        return zelf.handle_event(o)

    def handle_event(zelf, *args, **kwargs):
        logging.debug("< %s" % ", ".join([str(x) for x in args]))
        res = kernel.cmnds.dispatch(*args, **kwargs)
        if not res: res = kernel.ctrl.dispatch(*args, **kwargs)
        if res: return res
         
## TESTBOT

class TestBot(Bot):

    """ Bot used in unittests. """

    channel = "#test"

    def __init__(zelf, *args, **kwargs):
        Bot.__init__(zelf, *args, **kwargs)
        zelf.results = []

    def _raw(zelf, *args, **kwargs):
        zelf.results.append(args[0])

    def check(zelf, *args, **kwargs):
        if args[0] in zelf.results: return True
