# core/object.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2014 B.H.J Thate"

## IMPORTS

from core.utils import get_strace, smooth, pretty, get_plugname, get_name
from core.utils import short_date, j, rtime, a_time, nr_days, cdir
from core.utils import headertxt, txt_parse, list_files, error
from core.defines import hostname, port
from core import __version__

import threading
import hashlib
import logging
import fcntl
import types
import json
import time
import os

## BASE

class Object(dict):

    """ basic Object on which the rest of the program is based. """

    def __getattribute__(zelf, *args, **kwargs):
        name = args[0]
        if name == "url": return zelf.get_url("get")
        if name == "url_show": return zelf.get_url("show")
        return dict.__getattribute__(zelf, *args, **kwargs)
     
    def __getattr__(zelf, name):
        if name in zelf: return zelf[name]
        if name == "_ready": zelf["_ready"] = threading.Event()
        if name == "saved": zelf["saved"] = time.ctime(time.time())
        if name not in zelf: raise AttributeError(name)
        return zelf[name]

    def __setattr__(zelf, name, value): return dict.__setitem__(zelf, name, value)

    ## FLEET

    def get_bot(zelf, *args, **kwargs): return zelf._target

    def announce(zelf, *args, **kwargs):
        """ announce to all the running cores. """
        from core import fleet, kernel
        for bot in fleet: bot.announce(args[0])
        #thr = kernel.workers.put(bot.announce, args[0])
                    
    def search(zelf, *args, **kwargs):
        o = Object()
        for key, value in zelf.items():
            if args[0] in value: o[key] = value 

    def represent(zelf, *args, **kwargs):
        return ", ".join([str(getattr(zelf, x, None)) for x in zelf.get_keys() if getattr(zelf, x, None)])

    ## STATE

    def define(zelf, *args, **kwargs):
        """ set a attribute on this object. """
        name = args[0]
        value = args[1]
        zelf[name] = value

    def register(zelf, *args, **kwargs):
        """ register a value in a list on this object. """
        logging.info("register %s %s" % (args[0], args[1]))
        o = Object()
        o.name = args[0]
        o.func = args[1]
        o.plugname = get_plugname(2)
        if o.name not in zelf: zelf[o.name] = []
        zelf[o.name].append(o)

    def remove(zelf, *args, **kwargs):
        name = args[0]
        if name in zelf: del zelf[name]
        zelf[name] = []

    ## GETTERS

    def get_cmnd(zelf, *args, **kwargs):
        """ determine the command in the zelf.txt attribute, if present. """
        from core import kernel
        if "txt" in zelf and zelf.txt:
            val = zelf.txt.split()[0]
            if not val: return
            if "cc" in zelf:
                if val[0] != zelf.cc: return
                val = val[1:]
            try: cmnd = kernel.cmnds[val]
            except KeyError: 
                try: cmnd = kernel.ctrl[val]
                except KeyError: cmnd = None
            return cmnd

    def get_rest(zelf, *args, **kwargs):
        """ get the rest of the txt attribute. """
        try: cmnd, rest = zelf.txt.split(" ", 1) ; return rest
        except ValueError: return ""

    def get_args(zelf, *args, **kwargs):
        """ get the arguments of the txt attribute. """
        if "txt" in zelf: return zelf.txt.split()

    def get_days(zelf, *args, **kwargs):
        """ get the number of days relative to the object's creation time. """
        t1 = time.time()
        t2 = a_time(zelf.get_timed())
        if t2:
            time_diff = float(t1 - t2)
            return nr_days(time_diff)

    def get_timed(zelf, *args, **kwargs):
        """ retrieve the creation time of an object. """
        val = ""
        if "Date" in zelf: val = short_date(zelf.Date)
        elif "date" in zelf: val = short_date(zelf.date)
        elif "published" in zelf: val = short_date(zelf.published)
        elif "saved" in zelf: val = short_date(zelf.saved)
        elif "added" in zelf: val = short_date(zelf.added)
        return val

    def get_parsed(zelf, *args, **kwargs):
        """ parse the txt attribute. """
        rest = zelf.get_rest()
        return txt_parse(rest)
        
    def get_keys(zelf, *args, **kwargs):
        """ skip the unwanted keys e.g those that start with a "_". """
        for key in zelf.keys():
            k = str(key)
            if k.startswith("_"): continue
            yield key

    def get_obj(zelf, *args, **kwargs):
        """ cloned object, with only the proper keys used. """
        res = Object()
        for key in zelf.get_keys(): res[key] = zelf[key]
        return res

    def get_slice(zelf, *args, **kwargs):
        """ take a slice of the Object. """
        o = Object()
        arguments = args[0]
        for arg in arguments:
            try: o[arg] = zelf[arg]
            except KeyError: continue
        return o

    def get_filetime(zelf, *args, **kwargs):
        """ timestamp of related filename. """         
        return zelf._path.split(os.sep)[-1]

    def get_filedate(zelf, *args, **kwargs):
        """ timestamp of related filename. """         
        return zelf._path.split(os.sep)[-2:]

    def get_url(zelf, *args, **kwargs):
        """ url of the object's file so that it can be retrieved when API server is running. """
        import core.service.api as api
        fn = os.sep.join(zelf._path.split(os.sep)[-2:])
        if fn: return "http://%s:%s/%s/%s" % (hostname, port, args[0], fn)

    def get_root(zelf, *args, **kwargs):
        from core import kernel
        return kernel.cfg.workdir

    ## CHECKERS

    def check_wanted(zelf, *args, **kwargs):
        """ whether an object is desired. """
        want = args[0]
        for key, value in want.items():
           if key == "format": continue
           if key not in zelf: continue
           if value.startswith("-"): continue
           if value not in str(zelf[key]): return False
        return True

    def check_notwanted(zelf, *args, **kwargs):
        """ whether an object is not desired. """
        not_want = args[0]
        for key, value in not_want.items():
           if key == "format": continue
           if key not in zelf: continue
           if value in zelf[key]: return True
        return False

    ## INPUT

    def load(zelf, *args, **kwargs):
        """ load a JSON file into this object. """
        if args: path = args[0]
        else: path = zelf._path
        ondisk = zelf.read(path)
        fromdisk = json.loads(ondisk)
        if "full" in kwargs: zelf.update(fromdisk)
        elif "data" in fromdisk: zelf.update(fromdisk["data"])
        else: zelf.update(fromdisk)
        if "saved" in fromdisk: zelf.saved = fromdisk["saved"]
        zelf._path = path
        return zelf

    def read(zelf, *args, **kwargs):
        """ read the JSON file from disk. """
        path = args[0]
        try: f = open(path, "r")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## PERSISTENCE

    def to_disk(zelf, *args, **kwargs):
        """ create JSON ready to be saved to disk. """
        path = args[0]
        todisk = Object()
        todisk.data = zelf.get_obj()
        todisk.saved_from = get_plugname(2)
        todisk.create_type = get_name(zelf)
        todisk.saved = time.ctime(time.time())
        todisk.signature = todisk.data.make_signature()
        try: result = todisk.to_json(indent=2, ensure_ascii=False, sort_keys=True)
        except TypeError: raise NoJSON()
        return result

    def save(zelf, *args, **kwargs):
        """ save JSON to disk. """
        if not args: t = rtime()
        else: t = args[0]
        zelf.sync(j(zelf.get_root(), t))
        return t

    def sync(zelf, *args, **kwargs):
        """ sync JSON to disk. """
        try: path = args[0] ; zelf._path = path
        except IndexError: path = zelf._path
        logging.info("# %s" % path)
        d, fn = os.path.split(path)
        cdir(d)
        todisk = zelf.to_disk(path, **kwargs)
        datafile = open(os.path.abspath(path) + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % "%s characters" % len(todisk))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        return zelf

    ## OUTPUT

    def reply(zelf, *args, **kwargs):
        """ send reply to origin. """
        if "outer" in zelf: zelf.outer.write(str(args[0]) + "\n") ; zelf.outer.flush() ; return
        txt = args[0]
        if "channel" in zelf: zelf._target.say(zelf.channel, txt)
        else: zelf._target._raw(txt)

    def say(zelf, *args, **kwargs):
        """ output text through the _target attribute. """
        zelf._target.say(args[0], args[1])

    def ok(zelf, *args, **kwargs):
        """ signal ok. """
        if "_target" in zelf: zelf.reply("ok %s" % " ".join([str(a) for a in args]))

    def show(zelf):
        """ list of key,value pairs. """
        return ["%s=%s" % (a, zelf[a]) for a in zelf.get_keys()]

    ## ITERATORS

    def objects(zelf, *args, **kwargs):
        """ list of all object's. """
        if "time" in kwargs: desired_time = kwargs["time"]
        else: desired_time = ""
        objs = []
        path = zelf.get_root()
        logging.debug(path)
        for fnn in list_files(path, **kwargs):
            obj = Object().load(fnn, **kwargs)
            if args and args[0] not in obj: continue
            if desired_time and desired_time not in obj.get_timed(): continue
            if "deleted" in obj and obj.deleted: continue
            objs.append(obj)
        return objs

    def selected(zelf, *args, **kwargs):
        """ list of desired objects. """
        parsed = args[0]
        if kwargs.get("debug", None): logging.warn("> parse %s" % parsed.pretty())
        if "exclude" in kwargs: exclude = kwargs["exclude"]
        else: exclude = None
        if exclude in parsed.args: parsed.args.remove(exclude)
        if "time" in parsed.wanted: kwargs["time"] = parsed.wanted.time
        objs = []
        for obj in zelf.objects(**kwargs):
            if not obj.check_wanted(parsed.wanted): continue
            if not "full" in kwargs and not obj.selector(parsed.args): continue
            if obj.check_notwanted(parsed.not_wanted): continue
            objs.append(obj)
        return objs

    ## SELECTOR

    def selector(zelf, *args, **kwargs):
        """ see if this objects has the desired attributes. """        
        wanted = args[0]
        if not wanted: return True
        go = False
        for w in wanted:
           if zelf.get(w, ""): go = True ; break
        return go

    ## TIME RELATED

    def first(zelf, *args, **kwargs):
        """ return the first object where the key and/or value matches. """
        key = args[0]
        try: value = args[1]
        except IndexError: value = None
        result = None
        for o in zelf.objects():
            if key not in o: continue
            if value and o[key] != value: continue
            if not result: result = o
            if o.get("saved", "") < result.get("saved", ""): result = o
        return result

    def last(zelf, *args, **kwargs):
        """ return the last object where the key and/or value matches. """
        key = args[0]
        try: value = args[1]
        except IndexError: value = None
        result = None
        for o in zelf.objects():
            if key not in o: continue
            if value and value != o[key]: continue
            if not result: result = o
            if a_time(short_date(o.saved)) > a_time(short_date(result.saved)): result = o
        return result

    ## WAITERS

    def ready(zelf):
        """ signal to ready state. """
        logging.debug("> ready %s" % get_name(zelf))
        zelf._ready.set()

    def clear(zelf):
        """ clear the ready state. """
        logging.debug("> clear %s" % get_name(zelf))
        zelf._ready.clear()

    def wait(zelf, sec=180.0):
        """ wait for ready state. """
        logging.debug("> wait %s" % get_name(zelf))
        try: zelf._ready.wait(sec)
        except: pass

    ## HELPERS

    def to_json(zelf, *args, **kwargs):
        """ JSON string representation of this object. """
        return json.dumps(zelf.get_obj(*args, **kwargs), default=smooth, *args, **kwargs)

    def to_full(zelf, *args, **kwargs):
        """ full JSON dump of this object. """
        return json.dumps(zelf, default=smooth, indent=4)

    def make_signature(zelf, sig=None):
        """ signature of the data contained in this object. """
        return str(hashlib.sha1(bytes(str(zelf), "utf-8")).hexdigest())

    def pretty(zelf, *args, **kwargs):
        """ nice formatted JSON string of this object. """
        return json.dumps(zelf.get_obj(), indent=2, default=pretty, sort_keys=True, ensure_ascii=True)
